import pytest
from quixote import cleanup

from wcs.formdef import FormDef
from wcs.wf.register_comment import JournalEvolutionPart
from wcs.workflows import Workflow

from ..form_pages.test_all import create_user
from ..utilities import clean_temporary_pub, create_temporary_pub, get_app, login


def setup_module(module):
    cleanup()


def teardown_module(module):
    clean_temporary_pub()


@pytest.fixture
def pub():
    pub = create_temporary_pub()
    pub.cfg['language'] = {'language': 'en'}
    pub.cfg['identification'] = {'methods': ['password']}
    pub.write_cfg()
    return pub


def test_register_comment_migrate(pub):
    Workflow.wipe()

    workflow = Workflow(name='register message')
    st1 = workflow.add_status('Status1', 'st1')
    add_message = st1.add_action('register-comment')
    add_message.level = None
    add_message.comment = '<div class="errornotice">message</div>'
    workflow.store()

    workflow.migrate()
    assert workflow.possible_status[0].items[0].level == 'error'
    assert workflow.possible_status[0].items[0].comment == 'message'

    # check the migration is skipped if there's an extra class
    add_message.level = None
    add_message.comment = '<div class="errornotice blah">message</div>'
    workflow.store()
    workflow.migrate()
    assert not workflow.possible_status[0].items[0].level
    assert workflow.possible_status[0].items[0].comment == '<div class="errornotice blah">message</div>'


def test_register_comment_level(pub):
    FormDef.wipe()
    Workflow.wipe()

    workflow = Workflow(name='register message')
    st1 = workflow.add_status('Status1', 'st1')
    add_message = st1.add_action('register-comment')
    add_message.level = 'error'
    add_message.comment = 'hello\n\nworld'
    workflow.store()

    formdef = FormDef()
    formdef.name = 'foobar'
    formdef.workflow_id = workflow.id
    formdef.store()

    resp = get_app(pub).get('/foobar/')
    resp = resp.form.submit('submit')  # -> validation
    resp = resp.form.submit('submit').follow()  # -> done
    assert [x.text.strip() for x in resp.pyquery('#evolutions .errornotice p')] == ['hello', 'world']

    add_message.comment = '<p>hello\n\nworld</p>'
    workflow.store()

    resp = get_app(pub).get('/foobar/')
    resp = resp.form.submit('submit')  # -> validation
    resp = resp.form.submit('submit').follow()  # -> done
    assert [x.text.strip() for x in resp.pyquery('#evolutions .errornotice p')] == ['hello\n\nworld']


def test_register_comment_legacy_value(pub):
    FormDef.wipe()
    user = create_user(pub)

    formdef = FormDef()
    formdef.name = 'foobar'
    formdef.store()

    formdata = formdef.data_class()()
    formdata.just_created()
    formdata.user_id = user.id
    formdata.store()
    formdata.perform_workflow()
    legacy_part = JournalEvolutionPart(formdata, 'x', None, None)
    legacy_part.content = 'hello\n\nworld'
    formdata.evolution[-1].add_part(legacy_part)
    formdata.store()

    app = login(get_app(pub), username='foo', password='foo')
    resp = app.get(formdata.get_url())
    assert [x.text.strip() for x in resp.pyquery('#evolutions .msg p')] == ['hello', 'world']
