import pytest
import responses
from quixote import cleanup

from wcs.formdef import FormDef
from wcs.qommon.http_request import HTTPRequest
from wcs.wf.wscall import JournalWsCallErrorPart, WebserviceCallStatusItem

from ..utilities import clean_temporary_pub, create_temporary_pub


def setup_module(module):
    cleanup()


def teardown_module(module):
    clean_temporary_pub()


@pytest.fixture
def pub(request):
    pub = create_temporary_pub()
    pub.cfg['language'] = {'language': 'en'}
    pub.write_cfg()
    req = HTTPRequest(None, {'SERVER_NAME': 'example.net', 'SCRIPT_NAME': ''})
    req.response.filter = {}
    req._user = None
    pub._set_request(req)
    pub.set_config(req)
    return pub


def test_wscall_record_errors(pub):
    pub.user_class.wipe()
    FormDef.wipe()

    formdef = FormDef()
    formdef.name = 'test'
    formdef.fields = []
    formdef.store()
    formdef.data_class().wipe()

    formdata = formdef.data_class()()
    formdata.just_created()
    formdata.store()

    wscall = WebserviceCallStatusItem()
    wscall.url = 'http://test/'
    wscall.varname = 'varname'
    wscall.action_on_4xx = ':pass'
    wscall.record_errors = True

    # error with json data (stored as a string)
    with responses.RequestsMock() as rsps:
        rsps.get('http://test', status=404, json={'err': 1})
        wscall.perform(formdata)
        assert len([x for x in formdata.evolution[-1].parts if isinstance(x, JournalWsCallErrorPart)]) == 1
        assert formdata.evolution[-1].parts[-1].get_json_export_dict() == {
            'type': 'wscall-error',
            'summary': '404 Not Found',
            'label': None,
            'data': '{"err": 1}',
        }

    # error with bytes that can be stored as string
    formdata.evolution[-1].parts = []
    formdata.store()
    with responses.RequestsMock() as rsps:
        rsps.get('http://test', status=404, body=b'test bytes')
        wscall.perform(formdata)
        assert len([x for x in formdata.evolution[-1].parts if isinstance(x, JournalWsCallErrorPart)]) == 1
        assert formdata.evolution[-1].parts[-1].get_json_export_dict() == {
            'type': 'wscall-error',
            'summary': '404 Not Found',
            'label': None,
            'data': 'test bytes',
        }

    # error with bytes that cannot be stored as string
    formdata.evolution[-1].parts = []
    formdata.store()
    with responses.RequestsMock() as rsps:
        rsps.get('http://test', status=404, body=b'\xf1')
        wscall.perform(formdata)
        assert len([x for x in formdata.evolution[-1].parts if isinstance(x, JournalWsCallErrorPart)]) == 1
        assert formdata.evolution[-1].parts[-1].get_json_export_dict() == {
            'type': 'wscall-error',
            'summary': '404 Not Found',
            'label': None,
            'data_b64': '8Q==\n',
        }

    # error with payload that cannot be converted to JSON
    pub.loggederror_class.wipe()
    formdata.evolution[-1].parts = []
    formdata.store()
    wscall.method = 'POST'
    wscall.post_data = {'xxx': '=Ellipsis'}
    wscall.perform(formdata)
    assert pub.loggederror_class.count() == 1
    assert (
        formdata.workflow_data['varname_payload_error'] == 'Object of type ellipsis is not JSON serializable'
    )
    assert (
        pub.loggederror_class.select()[0].summary
        == 'Invalid payload (Object of type ellipsis is not JSON serializable)'
    )
